from dataclasses import dataclass
from typing import Type, Dict

@dataclass
class NludbRequest():
  pass

@dataclass
class NludbResponse():
  pass

  @staticmethod
  def safely_from_dict(d: any) -> Dict:
    """Last resort if subclass doesn't override: pass through."""
    return d




# from nludb.types.async_task import NludbTaskStatus
# import pytest
# from os import path
# from .helpers import _random_index, _random_name, _nludb

# from nludb import NLUDB, EmbeddingModels, EmbeddingIndex

# __author__ = "Edward Benson"
# __copyright__ = "Edward Benson"
# __license__ = "MIT"


# def test_file_upload():
#     nludb = _nludb()
#     name_a = "{}.mkd".format(_random_name())
#     a = nludb.upload_file(
#         name=name_a,
#         content="A"
#     )
#     assert(a.id is not None)
#     assert(a.name == name_a)
#     assert(a.type == 'mkd')

#     name_b = "{}.txt".format(_random_name())
#     b = nludb.upload_file(
#         name=name_b,
#         content="B"
#     )
#     assert(b.id is not None)
#     assert(b.name == name_b)
#     assert(b.type == 'txt')
#     assert(a.id != b.id)

#     a.delete()
#     b.delete()

#     test_filename = path.join(
#         path.dirname(path.realpath(__file__)),
#         'test_img.png'
#     )
#     with open(test_filename, "rb") as f:
#         name_c = "{}.png".format(_random_name())
#         c = nludb.upload_file(
#             name=name_b,
#             content=f
#         )
#         assert(c.id is not None)
#         assert(c.name == name_c)
#         assert(c.type == 'png')
#         c.delete()


# # def test_file_scrape():
# #     nludb = _nludb()

# #     name_a = "{}.html".format(_random_name())
# #     a = nludb.scrape_file(
# #         name=name_a,
# #         url="https://edwardbenson.com/2020/10/gpt3-travel-agent"
# #     )
# #     assert(a.id is not None)
# #     assert(a.name == name_a)
# #     assert(a.type == 'html')

# #     name_b = "{}.html".format(_random_name())
# #     b = nludb.scrape_file(
# #         name=name_b,
# #         url="https://edwardbenson.com/2018/09/case-of-the-murderous-ai"
# #     )
# #     assert(b.id is not None)    
# #     assert(a.id != b.id)
# #     assert(b.name == name_b)
# #     assert(b.type == 'html')

# #     a.delete()
# #     b.delete()

# # def test_file_upload_then_parse():
# #     nludb = _nludb()

# #     name_a = "{}.txt".format(_random_name())
# #     a = nludb.upload_file(
# #         name=name_a,
# #         content="This is a test."
# #     )
# #     assert(a.id is not None)

# #     q1 = a.query()
# #     assert(len(q1.blocks) == 0)

# #     task  = a.convert()
# #     task._run_development_mode()
# #     task.wait()

# #     q1 = a.query()
# #     assert(len(q1.blocks) == 2)
# #     assert(q1.blocks[0].type == 'doc')    
# #     assert(q1.blocks[1].type == 'paragraph')    
# #     assert(q1.blocks[1].value == 'This is a test.')

# #     name_b = "{}.mkd".format(_random_name())
# #     b = nludb.upload_file(
# #         name=name_a,
# #         content="""# Header

# # This is a test."""
# #     )
# #     assert(b.id is not None)

# #     q1 = b.query()
# #     assert(len(q1.blocks) == 0)

# #     task  = b.convert()
# #     task._run_development_mode()
# #     task.wait()

# #     q1 = b.query()
# #     assert(len(q1.blocks) == 3)
# #     assert(q1.blocks[0].type == 'doc')    
# #     assert(q1.blocks[2].type == 'paragraph')    
# #     assert(q1.blocks[2].value == 'This is a test.')
# #     assert(q1.blocks[1].type == 'h1')    
# #     assert(q1.blocks[1].value == 'Header')

# #     a.delete()
# #     b.delete()



import logging
import json
from typing import Union, List, Dict

from nludb import __version__
from nludb.api.base import ApiBase, AsyncTask
from nludb.types.classifier import ClassifierCreateRequest, ClassifierCreateResponse, ClassifyRequest, ClassifyResponse
from nludb.types.embedding_index import *

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

class Classifier:
  """A persistent, read-optimized index over embeddings.
  """

  def __init__(self, nludb: ApiBase, id: str = None, name: str = None, model: str = None, labels: List[str] = None):
    if id is None and model is None:
      raise Exception("Either an ID or a model must be provided")

    self.nludb = nludb
    self.name = name
    self.id = id
    self.model = model
    self.labels = labels

  @staticmethod
  def create(
    nludb: ApiBase,
    model: str,
    name: str = None,
    upsert: bool = True,
    save: bool = None,
    labels: List[str] = None
  ) -> "Classifier":
    if save == False:
      return Classifier(nludb, id=None, model=model, name=name, labels=labels)
    else:
      raise Exception("Persistent classifiers not yet supported.")
      req = ClassifierCreateRequest(
        model=model,
        name=name,
        upsert=upsert,
        save=save,
        labels=labels
      )
      res = nludb.post('classifier/create', req)
      return ClassifierCreateResponse(
        nludb=nludb,
        name=req.name,
        id=res.get("classifierId", None)
      )

  def classify(
    self, 
    docs: List[str],
    model: str = None,
    labels: List[str] = None,
    k: int = None
  ) -> ClassifyResponse:
    if self.id is None and self.model is None:
      raise Exception("Neither an ID nor a model was found on the classifier object. Please reinitialize with one or the other.")

    if self.id is None and (labels is None or len(labels) == 0) and (self.labels is None or len(self.labels) == 0):
      raise Exception('Since you are calling a stateless classifier, please include output labels in your classify request.')
    
    if self.id is not None and labels is not None and len(labels) > 0:
      raise Exception("Since you are calling a stateful classifier, you can not include in-line labels in your classify request. Please add them first.")
    
    req = ClassifyRequest(
      docs=docs,
      classifierId=self.id,
      model=model if model is not None else self.model,
      labels=labels if (labels is not None and len(labels) > 0) else self.labels,
      k=k
    )
    return self.nludb.post(
      'classifier/classify',
      req,
      expect=ClassifyResponse
    )


from typing import List
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponse

class ModelType:
  embedder = "embedder"
  parser = "parser"
  classifier = "classifier"

class ModelAdapterType:
  nludbDocker = "nludbDocker"
  nludbSagemaker = "nludbSagemaker"
  huggingface = "huggingface"
  openai = "openai"

class LimitUnit:
  words = "words"
  characters = "characters"
  bytes = "bytes"

@dataclass
class Model(NludbResponse):
  id: str = None
  name: str = None
  modelType: str = None
  url: str = None
  adapterType: str = None
  isPublic: bool = None
  handle: str = None
  description: str = None
  dimensionality: int = None
  limitAmount: int = None
  limitUnit: str = None
  apiKey: str = None
  metadata: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "Model":
    return Model(
      id = d.get('id', None),
      name = d.get('name', None),
      modelType = d.get('modelType', None),
      url = d.get('url', None),
      adapterType = d.get('adapterType', None),
      isPublic = d.get('isPublic', None),
      handle = d.get('handle', None),
      description = d.get('description', None),
      dimensionality = d.get('dimensionality', None),
      limitAmount = d.get('limitAmount', None),
      limitUnit = d.get('limitUnit', None),
      apiKey = d.get('apiKey', None),
      metadata = d.get('metadata', None)
    )

@dataclass
class CreateModelRequest(NludbRequest):
  id: str = None
  name: str = None
  modelType: str = None
  url: str = None
  adapterType: str = None
  isPublic: bool = None
  handle: str = None
  description: str = None
  dimensionality: int = None
  limitAmount: int = None
  limitUnit: str = None
  apiKey: str = None
  metadata: str = None
  upsert: bool = None

@dataclass
class DeleteModelRequest(NludbRequest):
  modelId: str

@dataclass
class ListPublicModelsRequest(NludbRequest):
  modelType: str

@dataclass
class ListPrivateModelsRequest(NludbRequest):
  modelType: str

@dataclass
class ListModelsResponse(NludbRequest):
  models: List[Model]

  @staticmethod
  def safely_from_dict(d: any) -> "ListModelsResponse":
    return ListModelsResponse(
      models = [Model.safely_from_dict(x) for x in d.get("models", [])]
    )



from nludb.types.async_task import NludbTaskStatus
import pytest
import os
import random
import string
import contextlib

from nludb import NLUDB, ClassifierModels, Classifier
from .helpers import _random_index, _random_name, _nludb

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

def test_classifier_create():
  nludb = _nludb()
  name = _random_name()

  # Should require name
  with pytest.raises(Exception):
    index = nludb.create_classifier(
      model=ClassifierModels.HF_ZERO_SHOT_LBART
    )

  # Should require model
  with pytest.raises(Exception):
    index = nludb.create_classifier(
      name="Test Index"
    )

  classifier = nludb.create_classifier(
    name="Test 2",
    model=ClassifierModels.HF_ZERO_SHOT_LBART,
    save=False
  )
  assert classifier is not None

  # Missing labels as a zero shot
  with pytest.raises(Exception):
      results = classifier.classify(docs=["Jurassic Park"])

  # Missing docs
  with pytest.raises(Exception):
      results = classifier.classify(labels=["Jurassic Park"])

  # Create one and let's use it
  classifier = nludb.create_classifier(
    name="Fruits",
    model=ClassifierModels.HF_ZERO_SHOT_LBART,
    save=False
  )
  assert classifier is not None
  r = classifier.classify(docs=["Banana"], labels=["Movie", "Food", "City"], k=2)
  assert len(r.hits) == 1
  assert len(r.hits[0]) == 2
  assert r.hits[0][0].value == "Food"
  assert r.hits[0][1].value == "City"

  r = classifier.classify(docs=["Banana"], labels=["Movie", "Food", "City"], k=3)
  
  assert len(r.hits) == 1
  assert len(r.hits[0]) == 3
  assert r.hits[0][0].value == "Food"
  assert r.hits[0][1].value == "City"
  assert r.hits[0][2].value == "Movie"

  r = classifier.classify(docs=["Banana", "Boston"], labels=["Movie", "Food", "City"], k=3)
  assert len(r.hits) == 2
  assert len(r.hits[0]) == 3
  assert r.hits[0][0].value == "Food"
  assert r.hits[0][1].value == "City"
  assert r.hits[0][2].value == "Movie"
  assert len(r.hits[1]) == 3
  assert r.hits[1][0].value == "City"
  assert r.hits[1][1].value == "Movie"
  assert r.hits[1][2].value == "Food"



import re
import os
from nludb import EmbeddingModels
from nludb import NLUDB 

FACT_TRIGGERS = [
  "fact", "rem", "remember", "learn", "true story"
]

FACT_PUNCT = [
  ":", " -- ", " that ", " - ", "--", "-", " "
]

class NludbChatbotBase:
  def __init__(
    self, 
    api_key: str, 
    index_name: str, 
    min_confidence: float = 0.4):
    self.nludb = NLUDB(api_key)
    self.index_name = index_name
    self.min_confidence = min_confidence
    self.index = self.nludb.create_index(
        index_name, 
        EmbeddingModels.QA, 
        upsert=True
    )

  def learn_fact(self, fact: str, externalId: str = None, externalType: str = None, metadata: dict = None):
    if fact is None:
      return None
    return self.index.insert(
      fact,
      externalId=externalId,
      externalType=externalType,
      metadata=metadata
    )

  def search_facts(self, query:  str) -> str:
    res = self.index.search(query)
    if res is None or res.hits is None or len(res.hits) == 0:
      return None
    hit = res.hits[0]
    if hit.score is None or hit.score < self.min_confidence:
      return None
    return hit.value 

  def extract_fact(self, input: str) -> str:
    lower = re.sub("\s\s+" , " ", input.lower())
    triggered = False
    for trigger in FACT_TRIGGERS:
      if lower.find(trigger) == 0:
        input = input[len(trigger):]
        triggered = True
        break
    if not triggered:
      return None
    
    # Now we strip the punctuation
    for punct in FACT_PUNCT:
      if input.find(punct) == 0:
        input = input[len(punct):]
        return input
    return input
  
  def dispatch(self, message: str, externalId: str = None, externalType: str = None, metadata: dict = None) -> str:
    maybe_fact = self.extract_fact(message)
    if maybe_fact is not None:
      self.learn_fact(maybe_fact, externalId=externalId, externalType=externalType, metadata=metadata)
      return "{}... OK, I got it!".format(maybe_fact)
    
    # Assume it's a question
    return self.search_facts(message)


def main():
  nludb_key = os.environ['NLUDB_KEY']
  if nludb_key is None:
    print("Please set the NLUDB_KEY environment variable")
    return

  index_name = os.environ.get('NLUDB_INDEX', "NludbChatbotBase Demo")
  bot = NludbChatbotBase(api_key=nludb_key, index_name=index_name)

  print('''NLUDB QA Bot Demonstration
  
Type "fact: <some fact>" to learn something new.
Type a question to retrieve knowledge.
  ''')
  while True:
    message = input("> ")   # Python 3
    response = bot.dispatch(message)
    if response:
      print(response)
    else:
      print("Unsure how to respond..")
    print("")

if __name__ == "__main__":
  main()


from typing import List
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponse
from nludb.types.search import Hit

@dataclass
class EmbedRequest(NludbRequest):
  texts: List[str]
  model: str

@dataclass
class EmbedResponse(NludbResponse):
  embeddings: List[List[float]]

  @staticmethod
  def safely_from_dict(d: any) -> "EmbedResponse":
    return EmbedResponse(
      embeddings = d.get('embeddings', None),
    )

@dataclass
class EmbedAndSearchRequest(NludbRequest):
  query: str
  docs: List[str]
  model: str
  k: int = 1

@dataclass
class EmbedAndSearchResponse(NludbRequest):
  hits: List[Hit] = None

  @staticmethod
  def safely_from_dict(d: any) -> "EmbedAndSearchResponse":
    hits = [Hit.safely_from_dict(h) for h in d.get("hits", [])]
    return EmbedAndSearchResponse(
      hits=hits
    )




from dataclasses import dataclass
from nludb.types.base import NludbResponse, NludbRequest

class NludbTaskStatus:
  waiting = "waiting"
  running = "running"
  succeeded = "succeeded"
  failed = "failed"

@dataclass
class TaskRunRequest(NludbRequest):
  taskId: str

@dataclass
class TaskStatusRequest(NludbRequest):
  taskId: str

@dataclass
class TaskStatusResponse(NludbResponse):
  taskId: str = None
  taskStatus: str = None
  taskCreatedOn: str = None
  taskLastModifiedOn: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "TaskStatusResponse":
    return TaskStatusResponse(
      taskId = d.get('taskId', None),
      taskStatus = d.get('taskStatus', None),
      taskCreatedOn = d.get('taskCreatedOn', None),
      taskLastModifiedOn = d.get('taskLastModifiedOn', None)
    )



import pytest
from nludb import EmbeddingModels
from .helpers import _random_index, _random_name, _nludb

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"


def test_basic_embeddings():
    nludb = _nludb()
    e1 = nludb.embed(["This is a test"], EmbeddingModels.QA)
    e1b = nludb.embed(["Banana"], EmbeddingModels.QA)
    assert(len(e1.embeddings) == 1)
    assert(len(e1.embeddings[0]) == 768)

    e2 = nludb.embed(["This is a test"], EmbeddingModels.QA)
    assert(len(e2.embeddings) == 1)
    assert(len(e2.embeddings[0]) == 768)

    e3 = nludb.embed(["This is a test"], EmbeddingModels.SIMILARITY)
    e3b = nludb.embed(["Banana"], EmbeddingModels.SIMILARITY)
    assert(len(e3.embeddings) == 1)
    assert(len(e3.embeddings[0]) == 768)

    e4 = nludb.embed(["This is a test"], EmbeddingModels.SIMILARITY)
    assert(len(e4.embeddings) == 1)
    assert(len(e4.embeddings[0]) == 768)

    e5 = nludb.embed(["This is a test"], EmbeddingModels.PARAPHRASE)
    e5b = nludb.embed(["Banana"], EmbeddingModels.PARAPHRASE)
    assert(len(e5.embeddings) == 1)
    assert(len(e5.embeddings[0]) == 768)

    e6 = nludb.embed(["This is a test"], EmbeddingModels.PARAPHRASE)
    assert(len(e6.embeddings) == 1)
    assert(len(e6.embeddings[0]) == 768)

    assert(e1.embeddings[0] == e2.embeddings[0])
    assert(e3.embeddings[0] == e4.embeddings[0])
    assert(e5.embeddings[0] == e6.embeddings[0])

    assert(e1.embeddings[0] != e1b.embeddings[0])
    assert(e3.embeddings[0] != e3b.embeddings[0])
    assert(e5.embeddings[0] != e5b.embeddings[0])

    assert(e1.embeddings[0] != e4.embeddings[0])
    assert(e1.embeddings[0] != e6.embeddings[0])
    assert(e3b.embeddings[0] != e5b.embeddings[0])
    assert(e4.embeddings[0] != e6.embeddings[0])

def test_basic_embedding_search():
    nludb = _nludb()
    docs = [
        "Armadillo shells are bulletproof.",
        "Dolphins sleep with one eye open.",
        "Alfred Hitchcock was frightened of eggs.",
        "Jonathan can help you with new employee onboarding",
        "The code for the New York office is 1234",
    ]
    query = "Who should I talk to about new employee setup?"
    results = nludb.embed_and_search(query, docs, EmbeddingModels.QA)
    assert(len(results.hits) == 1)
    assert(results.hits[0].value == "Jonathan can help you with new employee onboarding")



import requests
import logging
import time

from nludb import __version__
from nludb.types.base import NludbRequest, NludbResponse
from dataclasses import asdict
from typing import Type, TypeVar, Generic
from nludb.types.async_task import *

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

import typing

T = typing.TypeVar('T', bound=NludbResponse)

class AsyncTask(Generic[T]):
  """Encapsulates a unit of asynchronously performed work."""
  nludb: "ApiBase" = None
  taskId: str = None
  taskStatus: str = None
  taskCreatedOn: str = None
  taskLastModifiedOn: str = None

  def __init__(
    self, 
    nludb: "ApiBase" = None, 
    taskId: str = None, 
    taskStatus: str = None, 
    taskCreatedOn: str = None, 
    taskLastModifiedOn: str = None,
    eventualResultType: Type[NludbResponse] = None
    ):
    self.nludb =  nludb
    self.taskId = taskId
    self.taskStatus = taskStatus
    self.taskCreatedOn = taskCreatedOn
    self.taskLastModifiedOn = taskLastModifiedOn
    self.eventualResultType = eventualResultType

  def update(self, response: "AsyncTask"):
    """Incorporates a `TaskStatusResponse` into this object."""
    self.taskId = response.taskId
    self.taskStatus = response.taskStatus
    self.taskCreatedOn = response.taskCreatedOn
    self.taskLastModifiedOn = response.taskLastModifiedOn

  def check(self):
    """Retrieves and incorporates a fresh status update."""
    req = TaskStatusRequest(
      self.taskId
    )
    resp = self.nludb.post(
      'task/status',
      payload=req,
      expect=TaskStatusResponse,
      asynchronous=True
    )
    self.update(resp)
  
  def _run_development_mode(self):
    """Forces the task to run remotely (for unit testing; works only in development mode)."""
    return
    # req = TaskRunRequest(
    #   self.taskId
    # )
    # resp = self.nludb.post(
    #   'task/next',
    #   payload=req,
    #   expect=TaskStatusResponse,
    #   asynchronous=True
    # )
    # self.update(resp)

  def wait(self, max_timeout_s: float=60, retry_delay_s: float=1):
    """Polls and blocks until the task has succeeded or failed (or timeout reached)."""
    start = time.time()
    self.check()
    if self.taskStatus == NludbTaskStatus.succeeded or self.taskStatus == NludbTaskStatus.failed:
      return
    time.sleep(retry_delay_s)

    while time.time() - start < max_timeout_s:
      time.sleep(retry_delay_s)
      self.check()
      if self.taskStatus == NludbTaskStatus.succeeded or self.taskStatus == NludbTaskStatus.failed:
        return
    
class ApiBase:
  """Base class for API connectivity. 
  
  Separated primarily as a hack to prevent ciruclar imports.
  """
  def __init__(
    self, 
    api_key: str=None, 
    api_domain: str="https://api.nludb.com/",
    api_version: int=1):
    self.api_key = api_key
    self.api_domain = api_domain
    self.api_version = api_version
    self.endpoint = "{}/api/v{}".format(api_domain, api_version)
  
  T = TypeVar('T', bound=NludbResponse)

  def post(
    self, 
    operation: str, 
    payload: NludbRequest = None,
    file: None = None,
    expect: T = NludbResponse,
    asynchronous: bool = False
  ) -> T:
    """Post to the NLUDB API.

    All responses have the format:
       {
         data: <actual response>,
         error?: {
           reason: message
         }
       }
    
    For the Python client we return the contents of the `data`
    field if present, and we raise an exception if the `error`
    field is filled in.
    """
    if self.api_key is None:
      raise Exception("Please set your NLUDB API key.")

    url = "{}/{}".format(self.endpoint, operation)
    
    if file is not None:
      data = asdict(payload) if payload is not None else {}

      # Note: requests seems to have a bug passing boolean (and maybe numeric?)
      # values in the midst of multipart form data. You need to manually convert
      # it to a string; otherwise it will pass as False or True (with the capital),
      # which is not standard notation outside of Python.
      for key in data:
        if data[key] is False:
          data[key] = 'false'
        elif data[key] is True:
          data[key] = 'true'

      resp = requests.post(
        url,
        files={"file": file},
        data=data,
        headers = {"Authorization": "Bearer {}".format(self.api_key)}
      )
    else:
      resp = requests.post(
        url,
        json=asdict(payload) if payload is not None else None,
        headers = {"Authorization": "Bearer {}".format(self.api_key)}
      )
    j = resp.json()
    
    # Error response
    if 'reason' in j:
      raise Exception(j['reason'])

    # Non-asynchronous response
    if not asynchronous:
      if 'data' not in j:
        raise Exception('No data property was present on response')
      return expect.safely_from_dict(j['data'])

    # Expect an asynchronous response
    if 'status' not in j:
      if 'data' not in j:
        raise Exception('No data property was present on response')
      return expect.safely_from_dict(j['data'])

    # There is a status present
    task_resp = TaskStatusResponse.safely_from_dict(j['status'])
    task = AsyncTask(nludb=self)
    task.update(task_resp)
    return task



import json
from dataclasses import dataclass

@dataclass
class Hit:
  id: str = None
  index: int = None
  indexSource: str = None
  value: str = None
  score: float = None 
  externalId: str = None
  externalType: str = None
  metadata: any = None

  @staticmethod
  def safely_from_dict(d: any) -> "Hit":
    metadata = d.get("metadata", None)
    if metadata is not None:
      try:
        metadata = json.loads(metadata)
      except:
        pass

    return Hit(
      id=d.get("id", None),
      index=d.get("index", None),
      indexSource=d.get("indexSource", None),
      value=d.get("value", d.get("text", None)),
      score=d.get("score", None),
      externalId=d.get("externalId", None),
      externalType=d.get("externalType", None),
      metadata=metadata
    )


from typing import List
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponse

class FileUploadType:
  file = "file"
  url = "url"

@dataclass
class FileUploadRequest(NludbRequest):
  type: str
  name: str
  convert: bool = False

@dataclass
class FileUploadResponse(NludbResponse):
  fileId: str

  @staticmethod
  def safely_from_dict(d: any) -> "FileUploadResponse":
    return FileUploadResponse(
      fileId = d.get('fileId', None)
    )

@dataclass
class FileDeleteRequest(NludbRequest):
  fileId: str

@dataclass
class FileDeleteResponse(NludbResponse):
  fileId: str

  @staticmethod
  def safely_from_dict(d: any) -> "FileDeleteResponse":
    return FileDeleteResponse(
      fileId = d.get('fileId', None)
    )

@dataclass
class FileConvertRequest(NludbRequest):
  fileId: str

@dataclass
class FileConvertResponse(NludbResponse):
  fileId: str

  @staticmethod
  def safely_from_dict(d: any) -> "FileConvertResponse":
    return FileConvertResponse(
      fileId = d.get('fileId', None)
    )

@dataclass
class Block(NludbRequest):
  blockId: str
  type: str
  value: str

  @staticmethod
  def safely_from_dict(d: any) -> "Block":
    return Block(
      blockId = d.get('blockId', None),
      type = d.get('type', None),
      value = d.get('value', None)
    )

@dataclass
class FileQueryRequest(NludbRequest):
  fileId: str

@dataclass
class FileQueryResponse(NludbResponse):
  fileId: str
  blocks: List[Block]

  @staticmethod
  def safely_from_dict(d: any) -> "FileQueryResponse":
    return FileQueryResponse(
      fileId = d.get('fileId', None),
      blocks = [Block.safely_from_dict(block) for block in d.get('blocks', None)]
    )



from nludb.types.async_task import NludbTaskStatus
import pytest
import os
import random
import string
import contextlib

from nludb import NLUDB, EmbeddingModels, EmbeddingIndex
from .helpers import _random_index, _random_name, _nludb

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

def test_index_create():
    nludb = _nludb()
    name = _random_name()

    # Should require name
    with pytest.raises(Exception):
        index = nludb.create_index(
            model=EmbeddingModels.QA
        )

    # Should require model
    with pytest.raises(Exception):
        index = nludb.create_index(
            name="Test Index"
        )

    index = nludb.create_index(
        name=name,
        model=EmbeddingModels.QA,
        upsert=True
    )
    assert index is not None

    # Duplicate creation should fail with upsert=False
    with pytest.raises(Exception, match=r".*already exists.*"):
        index = nludb.create_index(
            name=name,
            model=EmbeddingModels.QA,
            upsert=False
        )

    index.delete()

def test_index_delete():
    nludb = _nludb()
    name = _random_name()
    index = nludb.create_index(
        name=name,
        model=EmbeddingModels.QA,
        upsert=True
    )
    assert(index.id is not None)

    index2 = nludb.create_index(
        name=name,
        model=EmbeddingModels.QA,
        upsert=True
    )
    assert(index.id == index2.id)
    
    index.delete()

    index3 = nludb.create_index(
        name=name,
        model=EmbeddingModels.QA,
        upsert=True
    )
    assert(index.id != index3.id)
    index3.delete()

def test_embed_task():
    nludb = _nludb()
    name = _random_name()
    with _random_index(nludb) as index:
        insert_results = index.insert("test", reindex=False )
        task = index.embed()

        assert (task.taskId is not None)
        assert (task.taskStatus is not None)
        assert (task.taskCreatedOn is not None)
        assert (task.taskLastModifiedOn is not None)
        assert (task.taskStatus == NludbTaskStatus.waiting)
        task._run_development_mode()
        task.wait()
        assert (task.taskStatus == NludbTaskStatus.succeeded)


def test_index_usage():
    nludb = _nludb()
    name = _random_name()

    with _random_index(nludb) as index:
        # Test for supressed reindexing
        A1 = "Ted can eat an entire block of cheese."
        Q1 = "Who can eat the most cheese"
        insert_results = index.insert(A1)
        search_results = index.search(Q1)

        # Now embed
        task = index.embed()
        task._run_development_mode()
        task.wait()
        task.check()
        assert (task.taskStatus == NludbTaskStatus.succeeded)

        search_results = index.search(Q1)
        assert(len(search_results.hits) == 1)
        assert(search_results.hits[0].value == A1)

        # Associate metadata
        A2 = "Armadillo shells are bulletproof."
        Q2 = "What is something interesting about Armadillos?"
        A2id = "A2id"
        A2type = "A2type"
        A2metadata = dict(
            id=A2id, 
            idid="{}{}".format(A2id, A2id),
            boolVal=True,
            intVal=123,
            floatVal=1.2
        )

        insert_results2 = index.insert(
            A2,
            externalId=A2id,
            externalType=A2type,
            metadata=A2metadata
        )
        search_results2 = index.search(Q2)
        assert(len(search_results2.hits) == 1)
        assert(search_results2.hits[0].value == A2)
        assert(search_results2.hits[0].externalId == None)
        assert(search_results2.hits[0].externalType == None)
        assert(search_results2.hits[0].metadata == None)

        search_results3 = index.search(Q2, includeMetadata=True)
        assert(len(search_results3.hits) == 1)
        assert(search_results3.hits[0].value == A2)
        assert(search_results3.hits[0].externalId == A2id)
        assert(search_results3.hits[0].externalType == A2type)

        assert(search_results3.hits[0].metadata == A2metadata)
        # Because I don't know pytest enough to fullly trust the dict comparison..
        assert(search_results3.hits[0].metadata["id"] == A2id)
        assert(search_results3.hits[0].metadata["idid"] == "{}{}".format(A2id, A2id))

        search_results4 = index.search(Q2, k=10)
        assert(len(search_results4.hits) == 2)
        assert(search_results4.hits[0].value == A2)
        assert(search_results4.hits[1].value == A1)



import logging
import json
from typing import Union, List, Dict

from nludb import __version__
from nludb.api.base import ApiBase, AsyncTask
from nludb.types.embedding_index import *

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

class EmbeddingIndex:
  """A persistent, read-optimized index over embeddings.
  """

  def __init__(self, nludb: ApiBase, id: str, name: str):
    self.nludb = nludb
    self.name = name
    self.id = id
  
  def insert(
    self, 
    value: str,
    externalId: str = None,
    externalType: str = None,
    metadata: Union[int, float, bool, str, List, Dict] = None,
    reindex: bool = True
  ) -> IndexInsertResponse:
    
    if isinstance(metadata, dict) or isinstance(metadata, list):
      metadata = json.dumps(metadata)

    req = IndexInsertRequest(
      self.id,
      value,
      externalId=externalId,
      externalType=externalType,
      metadata=metadata,
      reindex=reindex,
    )
    return self.nludb.post(
      'embedding-index/insert',
      req,
      expect=IndexInsertResponse
    )

  def embed(self) -> AsyncTask(IndexEmbedResponse):
    req = IndexEmbedRequest(
      self.id
    )
    return self.nludb.post(
      'embedding-index/embed',
      req,
      expect=IndexEmbedRequest,
      asynchronous=True
    )

  def delete(self) -> IndexDeleteResponse:
    req = IndexDeleteRequest(
      self.id
    )
    return self.nludb.post(
      'embedding-index/delete',
      req,
      expect=IndexDeleteResponse
    )

  def search(
    self, 
    query: str,
    k: int = 1,
    includeMetadata: bool = False
  ) -> IndexSearchResponse:
    req = IndexSearchRequest(
      self.id,
      query,
      k=k,
      includeMetadata=includeMetadata,
    )
    return self.nludb.post(
      'embedding-index/search',
      req,
      expect=IndexSearchResponse
    )

  @staticmethod
  def create(
    nludb: ApiBase,
    name: str,
    model: str,
    upsert: bool = True,
    externalId: str = None,
    externalType: str = None,
    metadata: any = None
  ) -> "EmbeddingIndex":
    req = IndexCreateRequest(
      name=name,
      model=model,
      upsert=upsert,
      externalId=externalId,
      externalType=externalType,
      metadata=metadata,
    )
    res = nludb.post('embedding-index/create', req)
    return EmbeddingIndex(
      nludb=nludb,
      name=req.name,
      id=res.get("id", None)
    )


import logging
import json
from typing import Union, List, Dict

from nludb import __version__
from nludb.api.base import ApiBase
from nludb.types.file import *

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

class File:
  """A file.
  """

  def __init__(self, nludb: ApiBase, id: str, name: str):
    self.nludb = nludb
    self.name = name
    self.id = id

  def delete(self) -> FileDeleteResponse:
    req = FileDeleteRequest(
      self.id
    )
    return self.nludb.post(
      'file/delete',
      req,
      expect=FileDeleteResponse
    )

  @staticmethod
  def upload(
    nludb: ApiBase,
    name: str,
    content: str) -> "File":
    req = FileUploadRequest(
      type=FileUploadType.file,
      name=name,
    )

    res = nludb.post(
      'file/upload',
      payload=req,
      file=(name, content, "multipart/form-data"),
      expect=FileUploadResponse
    )

    return File(
      nludb=nludb,
      name=req.name,
      id=res.fileId
    )

  @staticmethod
  def scrape(
    nludb: ApiBase,
    name: str,
    url: str) -> "File":
    req = FileUploadRequest(
      type=FileUploadType.url,
      name=name,
      url=url
    )

    res = nludb.post(
      'file/upload',
      payload=req,
      expect=FileUploadResponse
    )

    return File(
      nludb=nludb,
      name=req.name,
      id=res.fileId
    )

  def convert(self):
    req = FileConvertRequest(
      fileId=self.id
    )

    return self.nludb.post(
      'file/convert',
      payload=req,
      expect=FileConvertResponse,
      asynchronous=True
    )

  def query(self):
    req = FileQueryRequest(
      fileId=self.id
    )

    return self.nludb.post(
      'file/query',
      payload=req,
      expect=FileQueryResponse
    )



import logging
import json
from typing import Union, List, Dict

from nludb import __version__
from nludb.api.base import ApiBase, AsyncTask
from nludb.types.model import *

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

class Models:
  """A persistent, read-optimized index over embeddings.
  """

  def __init__(self, nludb: ApiBase):
    self.nludb = nludb
  
  def create(
    self, 
    name: str,
    description: str,
    modelType: str,
    url: str,
    adapterType: str,
    isPublic: bool,
    handle: str = None,
    dimensionality: int = None,
    limitAmount: int = None,
    limitUnit: str = None,
    apiKey: str = None,
    metadata: Union[str, Dict, List] = None,
    upsert: bool = None
  ) -> Model:
    
    if isinstance(metadata, dict) or isinstance(metadata, list):
      metadata = json.dumps(metadata)

    req = CreateModelRequest(
      name=name,
      modelType=modelType,
      url=url,
      adapterType=adapterType,
      isPublic=isPublic,
      handle=handle,
      description=description,
      dimensionality=dimensionality,
      limitAmount=limitAmount,
      limitUnit=limitUnit,
      apiKey=apiKey,
      metadata=metadata,
      upsert=upsert
    )
    return self.nludb.post(
      'model/create',
      req,
      expect=Model
    )

  def listPublic(
    self,
    modelType: str = None
  ) -> ListModelsResponse:
    return self.nludb.post(
      'model/public',
      ListPublicModelsRequest(modelType=modelType),
      expect=ListModelsResponse
    )

  def listPrivate(
    self,
    modelType = None
  ) -> ListModelsResponse:
    return self.nludb.post(
      'model/private',
      ListPrivateModelsRequest(modelType=modelType),
      expect=ListModelsResponse
    )

  def delete(
    self,
    id
  ) -> Model:
    return self.nludb.post(
      'model/delete',
      DeleteModelRequest(modelId=id),
      expect=Model
    )



import logging
from typing import Tuple, List

from nludb import __version__
from nludb.api.base import ApiBase
from nludb.types.embedding import EmbedRequest, EmbedResponse, EmbedAndSearchRequest, EmbedAndSearchResponse
from nludb.types.embedding_index import IndexCreateRequest
from nludb.embedding_index import EmbeddingIndex
from nludb.classifier import Classifier
from nludb.file import File
from nludb.types.embedding_models import EmbeddingModels
from nludb.types.file import FileUploadRequest, FileUploadResponse
from nludb.types.parsing import ParseRequest, ParseResponse, TokenMatcher, PhraseMatcher, DependencyMatcher
from nludb.types.parsing_models import ParsingModels
from nludb.models import Models

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

class NLUDB(ApiBase):
  """NLUDB Client Library.
  """
  def __init__(
    self, 
    api_key: str=None, 
    api_domain: str="https://api.nludb.com/",
    api_version: int=1):
    super().__init__(api_key, api_domain, api_version)
    self.models = Models(self)
 
  def create_index(
    self, 
    name: str,
    model: str,
    upsert: bool = True,
    externalId: str = None,
    externalType: str = None,
    metadata: any = None
  ) -> EmbeddingIndex:
    return EmbeddingIndex.create(
      nludb=self,
      name=name,
      model=model,
      upsert=upsert,
      externalId=externalId,
      externalType=externalType,
      metadata=metadata
    )

  def create_classifier(
    self, 
    name: str,
    model: str,
    upsert: bool = True,
    save: bool = None,
    labels: List[str] = None
  ) -> Classifier:
    return Classifier.create(
      nludb=self,
      model=model,
      name=name,
      upsert=upsert,
      save=save,
      labels=labels
    )

  def upload_file(
    self,
    name: str,
    content: str
  ) -> File:
    return File.upload(
      self,
      name,
      content
    )

  def scrape_file(
    self,
    name: str,
    url: str
  ) -> File:
    return File.upload(
      self,
      name,
      url
    )

  def embed(
    self, 
    texts: List[str],
    model: str
  ) -> EmbedResponse:
    req = EmbedRequest(
      texts=texts,
      model=model
    )
    return self.post(
      'embedding/create',
      req,
      expect=EmbedResponse
    )

  def embed_and_search(
    self, 
    query: str,
    docs: List[str],
    model: str,
    k: int = 1
  ) -> EmbedAndSearchResponse:
    req = EmbedAndSearchRequest(
      query=query,
      docs=docs,
      model=model,
      k=k
    )
    return self.post(
      'embedding/search',
      req,
      expect=EmbedAndSearchResponse
    )

  def parse(
    self,
    docs: List[str],
    model: str = ParsingModels.EN_DEFAULT,
    tokenMatchers: List[TokenMatcher] = None,
    phraseMatchers: List[PhraseMatcher] = None,
    dependencyMatchers: List[DependencyMatcher] = None,
  ) -> ParseResponse:
    req = ParseRequest(
      docs = docs,
      model = model,
      tokenMatchers = tokenMatchers,
      phraseMatchers = phraseMatchers,
      dependencyMatchers = dependencyMatchers
    )
    return self.post(
      'parser/parse',
      req,
      expect=ParseResponse
    )




from typing import List, Dict
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponse

@dataclass
class Token:
  text: str
  textWithWs: str
  whitespace: str
  head: str
  headI: int
  leftEdge: str
  rightEdge: str
  entType: str
  entIob: str
  lemma: str
  normalized: str
  shape: str
  prefix: str
  suffix: str
  isAlpha: bool
  isAscii: bool
  isDigit: bool
  isTitle: bool
  isPunct: bool
  isLeftPunct: bool
  isRightPunct: bool
  isSpace: bool
  isBracket: bool
  isQuote: bool
  isCurrency: bool
  likeUrl: bool
  likeNum: bool
  likeEmail: bool
  isOov: bool
  isStop: bool
  pos: str
  tag: str
  dep: str
  lang: str
  prob: float
  charIndex: int
  tokenIndex: int

  @staticmethod
  def safely_from_dict(d: any) -> "Token":
    return Token(
      text = d.get("text", None),
      textWithWs = d.get("textWithWs", None),
      whitespace = d.get("whitespace", None),
      head = d.get("head", None),
      headI = d.get("headI", None),
      leftEdge = d.get("leftEdge", None),
      rightEdge = d.get("rightEdge", None),
      entType = d.get("entType", None),
      entIob = d.get("entIob", None),
      lemma = d.get("lemma", None),
      normalized = d.get("normalized", None),
      shape = d.get("shape", None),
      prefix = d.get("prefix", None),
      suffix = d.get("suffix", None),
      isAlpha = d.get("isAlpha", None),
      isAscii = d.get("isAscii", None),
      isDigit = d.get("isDigit", None),
      isTitle = d.get("isTitle", None),
      isPunct = d.get("isPunct", None),
      isLeftPunct = d.get("isLeftPunct", None),
      isRightPunct = d.get("isRightPunct", None),
      isSpace = d.get("isSpace", None),
      isBracket = d.get("isBracket", None),
      isQuote = d.get("isQuote", None),
      isCurrency = d.get("isCurrency", None),
      likeUrl = d.get("likeUrl", None),
      likeNum = d.get("likeNum", None),
      likeEmail = d.get("likeEmail", None),
      isOov = d.get("isOov", None),
      isStop = d.get("isStop", None),
      pos = d.get("pos", None),
      tag = d.get("tag", None),
      dep = d.get("dep", None),
      lang = d.get("lang", None),
      prob = d.get("prob", None),
      charIndex = d.get("charIndex", None),
      tokenIndex = d.get("tokenIndex", None)
    )

  @staticmethod
  def from_spacy(t: any) -> "Token":
    return Token(
      text = t.text,
      textWithWs = t.text_with_ws,
      whitespace = t.whitespace_,
      head = t.head.text,
      headI = t.head.i,
      leftEdge = t.left_edge.text,
      rightEdge = t.right_edge.text,
      entType = t.ent_type_,
      entIob = t.ent_iob_,
      lemma = t.lemma_,
      normalized = t.norm_,
      shape = t.shape_,
      prefix = t.prefix_,
      suffix = t.suffix_,
      isAlpha = t.is_alpha,
      isAscii = t.is_ascii,
      isDigit = t.is_digit,
      isTitle = t.is_title,
      isPunct = t.is_punct,
      isLeftPunct = t.is_left_punct,
      isRightPunct = t.is_right_punct,
      isSpace = t.is_space,
      isBracket = t.is_bracket,
      isQuote = t.is_quote,
      isCurrency = t.is_currency,
      likeUrl = t.like_url,
      likeNum = t.like_num,
      likeEmail = t.like_email,
      isOov = t.is_oov,
      isStop = t.is_stop,
      pos = t.pos_,
      tag = t.tag_,
      dep = t.dep_.upper(),
      lang = t.lang_,
      prob = t.prob,
      charIndex = t.idx,
      tokenIndex = t.i,
    )

@dataclass
class Entity:
  text: str
  textWithWs: str
  startChar: int
  endChar: int
  startToken: int
  endToken: int
  label: str
  lemma: str

  @staticmethod
  def safely_from_dict(d: any) -> "Entity":
    return Entity(
      text = d.get("text", None),
      textWithWs = d.get("textWithWs", None),
      startChar = d.get("startChar", None),
      endChar = d.get("endChar", None),
      startToken = d.get("startToken", None),
      endToken = d.get("endToken", None),
      label = d.get("label", None),
      lemma = d.get("lemma", None)
    )

  @staticmethod
  def from_spacy(ent: any) -> "Entity":
    return Entity(
      text=ent.text,
      textWithWs=ent.text_with_ws,
      startChar=ent.start_char,
      endChar=ent.end_char,
      startToken=ent.start,
      endToken=ent.end,
      label=ent.label_,
      lemma=ent.lemma_,
    )

@dataclass
class Span:
  text: str
  textWithWs: str
  startChar: int
  endChar: int
  startToken: int
  endToken: int
  label: str
  lemma: str
  sentiment: float

  @staticmethod
  def safely_from_dict(d: any) -> "Span":
    return Span(
      text = d.get("text", None),
      textWithWs = d.get("textWithWs", None),
      startChar = d.get("startChar", None),
      endChar = d.get("endChar", None),
      startToken = d.get("startToken", None),
      endToken = d.get("endToken", None),
      label = d.get("label", None),
      lemma = d.get("lemma", None),
      sentiment = d.get("sentiment", None)
    )

  @staticmethod
  def from_spacy(ent: any) -> "Span":
    return Span(
      text=ent.text,
      textWithWs=ent.text_with_ws,
      startChar=ent.start_char,
      endChar=ent.end_char,
      startToken=ent.start,
      endToken=ent.end,
      label=ent.label_,
      lemma=ent.lemma_,
      sentiment=ent.sentiment
    )

@dataclass
class Sentence:
  text: str
  tokens: List[Token]
  entities: List[Entity]
  spans: List[Span]
  sentiment: float

  @staticmethod
  def safely_from_dict(d: any) -> "Sentence":
    tokens = [Token.safely_from_dict(h) for h in d.get("tokens", [])]
    entities = [Entity.safely_from_dict(h) for h in d.get("entities", [])]
    spans = [Span.safely_from_dict(h) for h in d.get("spans", [])]

    return Sentence(
      text=d.get("text", None),
      tokens=tokens,
      entities=entities,
      spans=spans,
      sentiment = d.get("sentiment", None)
    )

  @staticmethod
  def from_spacy(d: any) -> "Sentence":
    tokens = [Token.from_spacy(t) for t in d]
    entities = [Entity.from_spacy(e) for e in d.ents]

    return Sentence(
      text=d.text,
      tokens=tokens,
      entities=entities,
      spans=[],
      sentiment=d.sentiment
    )

@dataclass
class Doc:
  text: str
  sentences: List[Sentence]
  spans: List[Span]
  model: str
  lang: str
  sentiment: float

  @staticmethod
  def safely_from_dict(d: any) -> "Doc":
    sentences = [Sentence.safely_from_dict(h) for h in d.get("sentences", [])]
    spans = [Span.safely_from_dict(h) for h in d.get("spans", [])]
    return Doc(
      text=d.get("text", None),
      sentences=sentences,
      spans=spans,
      model=d.get("model", None),
      lang=d.get("lang", None),
      sentiment = d.get("sentiment", None)
    )

  @staticmethod
  def from_spacy(text: str, model: str, d: any) -> "Doc":
    sentences = [Sentence.from_spacy(s) for s in d.sents]
    spans = []
    for label in d.spans:
      span_group = d.spans[label]
      for s in span_group:
        spans.append(Span.safely_from_dict(s))

    ret = Doc(
      text=text,
      sentences=sentences,
      spans=spans,
      model=model,
      lang=d.lang_,
      sentiment=d.sentiment
    )
    return ret
  
@dataclass
class ParseResponse(NludbResponse):
  docs: List[Doc] = None

  @staticmethod
  def safely_from_dict(d: any) -> "ParseResponse":
    docs = [Doc.safely_from_dict(h) for h in d.get("docs", [])]
    return ParseResponse(
      docs=docs
    )

MatcherClause = Dict[str, any]
Matcher = List[MatcherClause]

@dataclass
class TokenMatcher():
  label: str = None
  patterns: List[Matcher] = None

  @staticmethod
  def safely_from_dict(d: any) -> "TokenMatcher":
    return TokenMatcher(
      label=d.get("label", None),
      patterns=d.get("patterns", [])
    )

@dataclass
class PhraseMatcher():
  label: str = None
  phrases: List[str] = None
  attr: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "PhraseMatcher":
    return PhraseMatcher(
      label=d.get("label", None),
      phrases=d.get("phrases", []),
      attr=d.get("attr", None)
    )

@dataclass
class DependencyMatcher():
  label: str = None
  # Note: Add a LABEL field to the matcher clause to have a token
  # get labeled in the response.
  patterns: List[Matcher] = None

  @staticmethod
  def safely_from_dict(d: any) -> "DependencyMatcher":
    return DependencyMatcher(
      label=d.get("label", None),
      patterns=d.get("patterns", [])
    )

@dataclass
class ParseRequest(NludbResponse):
  docs: List[str] = None
  model: str = None
  tokenMatchers: List[TokenMatcher] = None
  phraseMatchers: List[PhraseMatcher] = None
  dependencyMatchers: List[DependencyMatcher] = None

  @staticmethod
  def safely_from_dict(d: any) -> "ParseRequest":
    token_matchers = [TokenMatcher.safely_from_dict(h) for h in d.get("tokenMatchers", [])]
    phrase_matchers = [PhraseMatcher.safely_from_dict(h) for h in d.get("phraseMatchers", [])]
    dependency_matchers = [DependencyMatcher.safely_from_dict(h) for h in d.get("dependencyMatchers", [])]

    return ParseRequest(
      docs=d.get("docs", []),
      model=d.get("model", None),
      tokenMatchers=token_matchers,
      phraseMatchers=phrase_matchers,
      dependencyMatchers=dependency_matchers
    )




from nludb.types.async_task import NludbTaskStatus
import pytest
import os
import random
import string
import contextlib

from nludb import NLUDB, EmbeddingModels, EmbeddingIndex
from .helpers import _random_index, _random_name, _nludb

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

def test_connect():
    """Test basic connection"""
    nludb = _nludb()



import os
from typing import List
from nludb import EmbeddingModels
from nludb import NLUDB 

# Note: Numpy and SKLearn isn't in requirements.txt since it is only
#       used in this example!
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity 

class NludbAnomalyDetector:
  def __init__(self, api_key: str, model: str):
    self.nludb = NLUDB(api_key)
    self.model = model
    self.datapoints = []
    self.embeddings = []

  def add_datapoints(self, values: List[str]):
    res = self.nludb.embed(values, self.model)
    for value, embedding in zip(values, res.embeddings):
      self.datapoints.append(value)
      self.embeddings.append(np.array(embedding))
  
  def compute_mean(self) -> List[float]:
    return np.mean(self.embeddings, axis=0)

  def find_furthest(self) -> str:
    mean = self.compute_mean()
    sim = cosine_similarity(self.embeddings, [mean])
    # distances = self.embeddings - mean
    # norms = np.linalg.norm(distances, axis=1)    
    # print(norms)
    print(sim)
    furthest = np.argmin(sim)
    return self.datapoints[furthest]
    
def main():
  nludb_key = os.environ['NLUDB_KEY']
  if nludb_key is None:
    print("Please set the NLUDB_KEY environment variable")
    return

  def new_detector() -> NludbAnomalyDetector:
    return NludbAnomalyDetector(api_key=nludb_key, model=EmbeddingModels.SIMILARITY)

  detector = new_detector()
  print('''NLUDB Anomaly Detection Demonstration
  
Enter a list of items to consider, pressing enter after each.
Enter a blank line to find the one most unlike the others.
  ''')
  while True:
    message = input("> ")   # Python 3
    if len(message) == 0:
      print("Finding which one doesn't belong..")
      print("My guess: {}".format(detector.find_furthest()))
      detector = new_detector()
      print("")
    else:
      detector.add_datapoints([message])

if __name__ == "__main__":
  main()


from typing import List
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponse
from nludb.types.search import Hit

@dataclass
class ClassifierCreateRequest(NludbRequest):
  model: str
  name: str = None
  upsert: bool = True
  save: bool = True
  labels: List[str] = None

@dataclass
class ClassifierCreateResponse(NludbResponse):
  classifierId: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "ClassifierCreateResponse":
    return ClassifierCreateResponse(
      classifierId = d.get('classifierId', None),
    )

@dataclass
class ClassifyRequest(NludbRequest):
  docs: List[str]
  classifierId: str = None
  model: str = None
  labels: List[str] = None
  k: int = None

@dataclass
class ClassifyResponse(NludbResponse):
  classifierId: str = None
  model: str = None
  hits: List[List[Hit]] = None

  @staticmethod
  def safely_from_dict(d: any) -> "ClassifyResponse":
    hits = [[Hit.safely_from_dict(h) for h in innerList] for innerList in d.get("hits", [])]
    return ClassifyResponse(
      classifierId = d.get('classifierId', None),
      model = d.get('model', None),
      hits = hits
    )

@dataclass
class LabelInsertRequest(NludbRequest):
  value: str
  externalId: str = None
  externalType: str = None
  metadata: str = None

@dataclass
class LabelInsertResponse(NludbResponse):
  labelId: str
  value: str
  externalId: str = None
  externalType: str = None
  metadata: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "LabelInsertResponse":
    return LabelInsertResponse(
      labelId = d.get('labelId', None),
      value = d.get('value', None),
      externalId = d.get('externalId', None),
      externalType = d.get('externalType', None),
      metadata = d.get('metadata', None),
    )

@dataclass
class ClassifierDeleteResponse(NludbResponse):
  classifierId: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "ClassifierDeleteResponse":
    return ClassifierDeleteResponse(
      classifierId = d.get('classifierId', None),
    )



from nludb.types.async_task import NludbTaskStatus
import pytest
import os
import random
import string
import contextlib

from nludb import NLUDB, EmbeddingModels, EmbeddingIndex, File

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"


def _random_name() -> str:
    letters = string.digits + string.ascii_letters
    id =''.join(random.choice(letters) for i in range(10))
    return "test_{}".format(id)

@contextlib.contextmanager
def _random_index(nludb: NLUDB, model: str = EmbeddingModels.QA) -> EmbeddingIndex:
    index = nludb.create_index(
        name=_random_name(),
        model=model,
        upsert=True
    )
    yield index
    index.delete()  # or whatever you need to do at exit

@contextlib.contextmanager
def _random_file(nludb: NLUDB, content: str = "") -> File:
    file = nludb.create_file(
      name=_random_name(),
      contents=content
    )
    yield file
    file.delete()  # or whatever you need to do at exit


def _nludb() -> NLUDB:
    assert('NLUDB_KEY' in os.environ)
    NLUDB_DOMAIN = None
    if 'NLUDB_DOMAIN' in os.environ:
        NLUDB_DOMAIN = os.environ['NLUDB_DOMAIN']
    if NLUDB_DOMAIN is None:
        nludb = NLUDB(
            api_key = os.environ['NLUDB_KEY']
        )
    else:
        nludb = NLUDB(
            api_key = os.environ['NLUDB_KEY'],
            api_domain=NLUDB_DOMAIN
        )
    return nludb


import pytest
from nludb import EmbeddingModels
from .helpers import _random_index, _random_name, _nludb
from nludb.types.parsing import TokenMatcher, PhraseMatcher, DependencyMatcher

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"


def test_parsing():
    nludb = _nludb()
    resp = nludb.parse(["This is a test"])
    assert(len(resp.docs) == 1)
    d = resp.docs[0]
    assert(len(d.sentences) == 1)
    s = d.sentences[0]

    assert(s.text == "This is a test")
    assert(len(s.tokens) == 4)
    t = s.tokens[0]
    assert(t.lemma == "this")

def test_token_matcher():
    nludb = _nludb()

    a_matcher = TokenMatcher(
      label="A_MATCHER",
      patterns=[
        [{"LOWER": "a"}, {"LOWER": "matcher"}]
      ]
    )

    b_matcher = TokenMatcher(
      label="B_MATCHER",
      patterns=[
        [{"LOWER": "see"}],
        [{"LOWER": "if"}, {"LOWER": "a"}]
      ]
    )

    resp = nludb.parse(
      ["Let's see if a matcher works."], 
      tokenMatchers=[a_matcher, b_matcher]
    )
    assert(len(resp.docs) == 1)
    d = resp.docs[0]
    assert(len(d.sentences) == 1)
    s = d.sentences[0]
    assert(len(d.spans) == 3)
    ans = {
      "a matcher": a_matcher.label,
      "see": b_matcher.label,
      "if a": b_matcher.label
    }
    for sp in d.spans:
      assert(sp.text in ans)
      assert(sp.label == ans[sp.text])

def test_phrase_matcher():
    nludb = _nludb()

    a_matcher = PhraseMatcher(
      label="A_MATCHER",
      phrases=["a matcher"]
    )

    b_matcher = PhraseMatcher(
      label="B_MATCHER",
      phrases=["if a", "see"]
    )

    resp = nludb.parse(
      ["Let's see if a matcher works."], 
      phraseMatchers=[a_matcher, b_matcher]
    )
    assert(len(resp.docs) == 1)
    d = resp.docs[0]
    assert(len(d.sentences) == 1)
    s = d.sentences[0]
    assert(len(d.spans) == 3)
    ans = {
      "a matcher": a_matcher.label,
      "see": b_matcher.label,
      "if a": b_matcher.label
    }
    for sp in d.spans:
      assert(sp.text in ans)
      assert(sp.label == ans[sp.text])

def test_phrase_matcher_attr():
    nludb = _nludb()

    a_matcher = PhraseMatcher(
      label="IP_ADDR",
      phrases=["127.0.0.1", "127.127.0.0"],
      attr="SHAPE"
    )

    resp = nludb.parse(
      ["Often the router will have an IP address such as 192.168.1.1 or 192.168.2.1."],
      phraseMatchers=[a_matcher]
    )
    assert(len(resp.docs) == 1)
    d = resp.docs[0]
    assert(len(d.sentences) == 1)
    s = d.sentences[0]
    assert(len(d.spans) == 2)
    ans = {
      "192.168.1.1": a_matcher.label,
      "192.168.2.1": a_matcher.label
    }
    for sp in d.spans:
      assert(sp.text in ans)
      assert(sp.label == ans[sp.text])

def test_token_phrase_matcher_combo():
    nludb = _nludb()

    a_matcher = PhraseMatcher(
      label="A_MATCHER",
      phrases=["a Matcher"]
    )

    b_matcher = PhraseMatcher(
      label="B_MATCHER",
      phrases=["if a", "see"]
    )

    c_matcher = PhraseMatcher(
      label="IP_ADDR",
      phrases=["127.0.0.1", "127.127.0.0", "88.88.88.88"],
      attr="SHAPE"
    )

    d_matcher = TokenMatcher(
      label="FOO",
      patterns = [[{"LOWER": "can"}]]
    )

    e_matcher = PhraseMatcher(
      label="CAP",
      phrases=["Ted"],
      attr="SHAPE"
    )

    f_matcher = PhraseMatcher(
      label="accordingly",
      phrases=["accordingly"],
      attr="LEMMA"
    )

    resp = nludb.parse(
      ["Let's see if a Matcher can match 44.33.22.11 accordingly."], 
      tokenMatchers=[d_matcher],
      phraseMatchers=[a_matcher, b_matcher, c_matcher, e_matcher, f_matcher]
    )

    assert(len(resp.docs) == 1)
    d = resp.docs[0]
    assert(len(d.sentences) == 1)
    s = d.sentences[0]
    assert(len(d.spans) == 7)
    ans = {
      "a Matcher": a_matcher.label,
      "see": b_matcher.label,
      "if a": b_matcher.label,
      "can": d_matcher.label,
      "44.33.22.11": c_matcher.label,
      "Matcher": e_matcher.label,
      "f_matcher": f_matcher.label,
      "Let": e_matcher.label,
      "accordingly": f_matcher.label
    }
    for sp in d.spans:
      assert(sp.text in ans)
      assert(sp.label == ans[sp.text])

def test_dependency_matcher():
    nludb = _nludb()

    a_matcher = DependencyMatcher(
      label="FOUNDED",
      patterns=[
        [
          {
            "RIGHT_ID": "anchor_founded",       # unique name
            "RIGHT_ATTRS": {"ORTH": "founded"}  # token pattern for "founded"
          }
        ]
      ]
    )

    resp = nludb.parse(
      ["Smith founded two companies"],
      dependencyMatchers=[a_matcher]
    )
    assert(len(resp.docs) == 1)
    d = resp.docs[0]
    assert(len(d.sentences) == 1)
    s = d.sentences[0]
    assert(len(d.spans) == 1)
    ans = {
      "founded": a_matcher.label
    }
    for sp in d.spans:
      assert(sp.text in ans)
      assert(sp.label == ans[sp.text])
    
    a_matcher = DependencyMatcher(
      label="FOUNDED",
      patterns=[
        [
          {
            "RIGHT_ID": "anchor_founded",       # unique name
            "RIGHT_ATTRS": {"ORTH": "founded"},  # token pattern for "founded"
            "LABEL": "FOUND"
          }
        ]
      ]
    )

    resp = nludb.parse(
      ["Smith founded two companies"],
      dependencyMatchers=[a_matcher]
    )
    assert(len(resp.docs) == 1)
    d = resp.docs[0]
    assert(len(d.sentences) == 1)
    s = d.sentences[0]
    assert(len(d.spans) == 2)
    ans = {
      "founded": a_matcher.label
    }
    for sp in d.spans:
      assert(sp.text in ans)
      assert(sp.label in ["FOUNDED", "FOUND"])


    pattern = [
        {
            "RIGHT_ID": "anchor_founded",
            "RIGHT_ATTRS": {"ORTH": "founded"},
            "LABEL": "ROOT"
        },
        {
            "LEFT_ID": "anchor_founded",
            "REL_OP": ">",
            "RIGHT_ID": "founded_subject",
            "RIGHT_ATTRS": {"DEP": "nsubj"},
            "LABEL": "WHO"
        },
        {
            "LEFT_ID": "anchor_founded",
            "REL_OP": ">",
            "RIGHT_ID": "founded_object",
            "RIGHT_ATTRS": {"DEP": "dobj"},
            "LABEL": "WHAT"
        },
        {
            "LEFT_ID": "founded_object",
            "REL_OP": ">",
            "RIGHT_ID": "founded_object_modifier",
            "RIGHT_ATTRS": {"DEP": {"IN": ["amod", "compound"]}},
        }
    ]
    sent = "Lee, an experienced CEO, has founded two AI startups."

    a_matcher = DependencyMatcher(
      label="FOUNDED",
      patterns=[pattern]
    )
    resp = nludb.parse(
      [sent],
      dependencyMatchers=[a_matcher]
    )

    assert(len(resp.docs) == 1)
    d = resp.docs[0]
    assert(len(d.sentences) == 1)
    s = d.sentences[0]
    assert(len(d.spans) == 4)
    spans = d.spans
    ans = {
      "startups": "WHAT",
      "Lee": "WHO",
      "founded": "ROOT",
      "Lee, an experienced CEO, has founded two AI startups": "FOUNDED"
    }
    assert(len(spans) == 4)
    for sp in spans:
      assert(sp.text in ans)
      assert(sp.label == ans[sp.text])





from typing import List
from dataclasses import dataclass
from nludb.types.base import NludbRequest, NludbResponse
from nludb.types.search import Hit
@dataclass
class IndexCreateRequest(NludbRequest):
  name: str
  model: str
  upsert: bool = True
  externalId: str = None
  externalType: str = None
  metadata: any = None

@dataclass
class IndexCreateResponse(NludbResponse):
  id: str

@dataclass
class IndexInsertRequest(NludbRequest):
  indexId: str
  value: str
  externalId: str = None
  externalType: str = None
  metadata: any = None
  reindex: bool = True

@dataclass
class IndexItemId(NludbResponse):
  indexId: str = None
  id: str = None

  @staticmethod
  def safely_from_dict(d: any) -> "IndexItemId":
    return IndexItemId(
      indexId = d.get('indexId', None),
      id = d.get('id', None)
    )

@dataclass
class IndexInsertResponse(NludbResponse):
  itemIds: List[IndexItemId] = None

  @staticmethod
  def safely_from_dict(d: any) -> "IndexInsertResponse":
    return IndexInsertResponse(
      itemIds = [IndexItemId.safely_from_dict(x) for x in d.get('itemIds', [])]
    )

@dataclass
class IndexEmbedRequest(NludbRequest):
  indexId: str

@dataclass
class IndexEmbedResponse(NludbResponse):
  indexId: str

  @staticmethod
  def safely_from_dict(d: any) -> "IndexEmbedResponse":
    return IndexEmbedResponse(
      indexId = d.get('indexId', None)
    )


@dataclass
class IndexDeleteRequest(NludbRequest):
  indexId: str

@dataclass
class IndexDeleteResponse(NludbResponse):
  indexId: str

  @staticmethod
  def safely_from_dict(d: any) -> "IndexDeleteResponse":
    return IndexDeleteResponse(
      indexId = d.get('indexId', None)
    )

@dataclass
class IndexSearchRequest(NludbRequest):
  indexId: str
  query: str
  k: int = 1
  includeMetadata: bool = False

@dataclass
class IndexSearchResponse(NludbResponse):
  hits: List[Hit] = None

  @staticmethod
  def safely_from_dict(d: any) -> "IndexSearchResponse":
    hits = [Hit.safely_from_dict(h) for h in d.get("hits", [])]
    return IndexSearchResponse(
      hits=hits
    )



from nludb.types.async_task import NludbTaskStatus
from nludb.types.model import ModelAdapterType
import pytest
import os
import random
import string
import contextlib

from nludb import NLUDB, ModelType

from .helpers import _random_index, _random_name, _nludb

__author__ = "Edward Benson"
__copyright__ = "Edward Benson"
__license__ = "MIT"

def test_model_create():
    nludb = _nludb()
    name = _random_name()

    my_models = nludb.models.listPrivate()
    orig_count = len(my_models.models)

    # Should require name
    with pytest.raises(Exception):
        index = nludb.models.create(
            description = "This is just for test",
            modelType = ModelType.embedder,
            url = "http://foo",
            adapterType = ModelAdapterType.nludbDocker,
            isPublic = True
        )

    # Should require description
    with pytest.raises(Exception):
        index = nludb.models.create(
            name = "Test Model",
            modelType = ModelType.embedder,
            url = "http://foo",
            adapterType = ModelAdapterType.nludbDocker,
            isPublic = True
        )

    # Should require model type
    with pytest.raises(Exception):
        index = nludb.models.create(
            name = "Test Model",
            description = "This is just for test",
            url = "http://foo",
            adapterType = ModelAdapterType.nludbDocker,
            isPublic = True
        )

    # Should require url
    with pytest.raises(Exception):
        index = nludb.models.create(
            name = "Test Model",
            description = "This is just for test",
            modelType = ModelType.embedder,
            adapterType = ModelAdapterType.nludbDocker,
            isPublic = True
        )

    # Should require adapter type
    with pytest.raises(Exception):
        index = nludb.models.create(
            name = "Test Model",
            description = "This is just for test",
            modelType = ModelType.embedder,
            url = "http://foo",
            isPublic = True
        )

    # Should require is public
    with pytest.raises(Exception):
        index = nludb.models.create(
            name = "Test Model",
            description = "This is just for test",
            modelType = ModelType.embedder,
            url = "http://foo",
            adapterType = ModelAdapterType.nludbDocker,
        )

    my_models = nludb.models.listPrivate()
    assert(len(my_models.models) == orig_count)

    model = nludb.models.create(
        name = _random_name(),
        description = "This is just for test",
        modelType = ModelType.embedder,
        url = "http://foo",
        adapterType = ModelAdapterType.nludbDocker,
        isPublic = False
    )

    my_models = nludb.models.listPrivate()
    assert(len(my_models.models) == orig_count+1)

    # No upsert doesn't work
    with pytest.raises(Exception):
      model = nludb.models.create(
          name = model.name,
          description = "This is just for test",
          modelType = ModelType.embedder,
          url = "http://foo",
          adapterType = ModelAdapterType.nludbDocker,
          isPublic = False
      )

    # Upsert does work
    model2 = nludb.models.create(
        name = model.name,
        description = "This is just for test 2",
        modelType = ModelType.embedder,
        url = "http://foo",
        adapterType = ModelAdapterType.nludbDocker,
        isPublic = False,
        upsert = True
    )

    assert(model2.id == model.id)

    my_models = nludb.models.listPrivate()
    assert(len(my_models.models) == orig_count+1)
    assert(my_models.models[0].id == model2.id)
    assert(my_models.models[0].description == model2.description)
    # assert(my_models.models[0].description != model.description)

    nludb.models.delete(model.id)

    my_models = nludb.models.listPrivate()
    assert(len(my_models.models) == orig_count)

def test_model_public():
    nludb = _nludb()
    name = _random_name()

    my_models = nludb.models.listPublic()
    orig_count = len(my_models.models)
    assert(len(my_models.models) > 0, True)

    # Make sure they can't be deleted.
    with pytest.raises(Exception):    
      nludb.models.delete(my_models.models[0].id)
    
